import datetime
import logging
from typing import List, Optional

from internetarchive import ArchiveSession, Search

from wikiteam3.dumpgenerator.config import Config

IA_MAX_RETRY = 5
logger = logging.getLogger(__name__)

def search_ia(apiurl: Optional[str] = None, indexurl: Optional[str] = None, addeddate_intervals: Optional[List[str]] = None):
    if apiurl is None:
        apiurl = 'api.php'.join(indexurl.rsplit('index.php', 1)) if indexurl else None
    if indexurl is None:
        indexurl = 'index.php'.join(apiurl.rsplit('api.php', 1)) if apiurl else None

    if not (apiurl and indexurl):
        raise ValueError('apiurl or indexurl must be provided')

    ia_session = ArchiveSession()

    query = f'(originalurl:"{apiurl}" OR originalurl:"{indexurl}")'
    if addeddate_intervals:
        query += f' AND addeddate:[{addeddate_intervals[0]} TO {addeddate_intervals[1]}]'
    search = Search(ia_session, query=query,
                    fields=['identifier', 'addeddate', 'title', 'subject', 'originalurl', 'uploader', 'item_size'],
                    sorts=['addeddate desc'], # newest first
                    max_retries=IA_MAX_RETRY, # default 5
                    )
    item = None
    for result in search: # only get the first result
        # {'identifier': 'wiki-wikiothingxyz-20230315',
        # 'addeddate': '2023-03-15T01:42:12Z',
        # 'subject': ['wiki', 'wikiteam', 'MediaWiki', .....]}
        if apiurl.lower() == result['originalurl'].lower() or indexurl.lower() == result['originalurl'].lower():
            logger.info(f'Original URL match: {result}')
            yield result
            item = result
        else:
            logger.warning(f'Original URL mismatch: {result}')

    if item is None:
        logger.warning('No suitable dump found at Internet Archive')
        return # skip


def search_ia_recent(config: Config, days: int = 365):

    now_utc = datetime.datetime.utcnow()
    now_utc_iso = now_utc.strftime("%Y-%m-%dT%H:%M:%SZ")

    one_year_ago = now_utc - datetime.timedelta(days=days)
    one_year_ago_iso = one_year_ago.strftime("%Y-%m-%dT%H:%M:%SZ")

    addeddate_intervals = [one_year_ago_iso, now_utc_iso]

    for item in search_ia(apiurl=config.api, indexurl=config.index, addeddate_intervals=addeddate_intervals):
        yield item


def any_recent_ia_item_exists(config: Config, days: int = 365):
    for item in search_ia_recent(config=config, days=days):
        print('Found an existing dump at Internet Archive')
        print(item)
        print(f'https://archive.org/details/{item["identifier"]}')
        return True

    return False


def search_ia_all(config: Config):
    for item in search_ia(apiurl=config.api, indexurl=config.index):
        yield item



import json
import re
import sys
from typing import Tuple

import requests

from wikiteam3.dumpgenerator.exceptions import ExportAbortedError, PageMissingError
from wikiteam3.dumpgenerator.log import log_error
from wikiteam3.dumpgenerator.dump.page.xmlexport.page_xml import get_XML_page
from wikiteam3.dumpgenerator.config import Config

def getXMLHeader(config: Config=None, session=None) -> Tuple[str, Config]:
    """Retrieve a random page to extract XML headers (namespace info, etc)"""
    # get the header of a random page, to attach it in the complete XML backup
    # similar to: <mediawiki xmlns="http://www.mediawiki.org/xml/export-0.3/"
    # xmlns:x....
    randomtitle = "Main_Page"  # previously AMF5LKE43MNFGHKSDMRTJ
    print(config.api)
    xml = ""
    disableSpecialExport = config.xmlrevisions or config.xmlapiexport
    if disableSpecialExport and config.api and config.api.endswith("api.php"):
        try:
            print("Getting the XML header from the API")
            # Export and exportnowrap exist from MediaWiki 1.15, allpages from 1.8
            r = session.get(
                config.api
                + "?action=query&export=1&exportnowrap=1&list=allpages&aplimit=1",
                timeout=10,
            )
            xml: str = r.text
            # Otherwise try without exportnowrap, e.g. Wikia returns a blank page on 1.19
            if not re.match(r"\s*<mediawiki", xml):
                r = session.get(
                    config.api
                    + "?action=query&export=1&list=allpages&aplimit=1&format=json",
                    timeout=10,
                )
                try:
                    xml = r.json()["query"]["export"]["*"]
                except KeyError:
                    pass
            if not re.match(r"\s*<mediawiki", xml):
                # Do without a generator, use our usual trick of a random page title
                r = session.get(
                    config.api
                    + "?action=query&export=1&exportnowrap=1&titles="
                    + randomtitle,
                    timeout=10,
                )
                xml = str(r.text)
            # Again try without exportnowrap
            if not re.match(r"\s*<mediawiki", xml):
                r = session.get(
                    config.api
                    + "?action=query&export=1&format=json&titles="
                    + randomtitle,
                    timeout=10,
                )
                try:
                    xml = r.json()["query"]["export"]["*"]
                except KeyError:
                    pass
        except requests.exceptions.RetryError:
            pass

    else:
        try:
            xml = "".join(
                [
                    x
                    for x in get_XML_page(
                        config=config, title=randomtitle, verbose=False, session=session
                    )
                ]
            )
        except PageMissingError as pme:
            # The <page> does not exist. Not a problem, if we get the <siteinfo>.
            xml = pme.xml
        # Issue 26: Account for missing "Special" namespace.
        # Hope the canonical special name has not been removed.
        # http://albens73.fr/wiki/api.php?action=query&meta=siteinfo&siprop=namespacealiases
        except ExportAbortedError:
            try:
                if config.api:
                    print("Trying the local name for the Special namespace instead")
                    r = session.get(
                        url=config.api,
                        params={
                            "action": "query",
                            "meta": "siteinfo",
                            "siprop": "namespaces",
                            "format": "json",
                        },
                        timeout=120,
                    )
                    config.export = (
                        json.loads(r.text)["query"]["namespaces"]["-1"]["*"] + ":Export"
                    )
                    xml = "".join(
                        [
                            x
                            for x in get_XML_page(
                                config=config,
                                title=randomtitle,
                                verbose=False,
                                session=session,
                            )
                        ]
                    )
            except PageMissingError as pme:
                xml = pme.xml
            except ExportAbortedError:
                pass

    header = xml.split("</mediawiki>")[0]
    if not re.match(r"\s*<mediawiki", xml):
        if config.xmlrevisions:
            # Try again the old way
            print(
                "Export test via the API failed. Wiki too old? Trying without xmlrevisions."
            )
            config.xmlrevisions = False
            header, config = getXMLHeader(config=config, session=session)
        else:
            print(xml)
            print("XML export on this wiki is broken, quitting.")
            log_error(config=config, to_stdout=True,
                      text="XML export on this wiki is broken, quitting.")
            sys.exit(1)
    return header, config



import requests


def get_JSON(request: requests.Response):
    """Strip Unicode BOM"""
    if request.text.startswith("\ufeff"):
        request.encoding = "utf-8-sig"
    # request.encoding = request.apparent_encoding
    try:
        return request.json()
    except:
        # Maybe an older API version which did not return correct JSON
        print("Error: Could not parse JSON")
        return {}



import re

import requests


def checkIndex(index="", cookies="", session: requests.Session=None):
    """Checking index.php availability"""
    r = session.post(url=index, data={"title": "Special:Version"}, timeout=30)
    if r.status_code >= 400:
        print(f"ERROR: The wiki returned status code HTTP {r.status_code}")
        return False
    raw = str(r.text)
    print("Checking index.php...", index)
    # Workaround for issue 71
    if (
        re.search(
            '(Special:Badtitle</a>|class="permissions-errors"|"wgCanonicalSpecialPageName":"Badtitle"|Login Required</h1>)',
            raw,
        )
        and not cookies
    ):
        print("ERROR: This wiki requires login and we are not authenticated")
        return False
    if re.search(
        '(page-Index_php|"wgPageName":"Index.php"|"firstHeading"><span dir="auto">Index.php</span>)',
        raw,
    ):
        print("Looks like the page called Index.php, not index.php itself")
        return False
    if re.search(
        '(This wiki is powered by|<h2 id="mw-version-license">|meta name="generator" content="MediaWiki|class="mediawiki)',
        raw,
    ):
        return True
    return False



import contextlib
import os
import re
import subprocess
import sys
import traceback
from typing import Dict

from file_read_backwards import FileReadBackwards

from wikiteam3.dumpgenerator.config import load_config, save_config
from wikiteam3.dumpgenerator.config import Config
from wikiteam3.dumpgenerator.cli import get_parameters, bye, welcome
from wikiteam3.dumpgenerator.dump.image.image import Image
from wikiteam3.dumpgenerator.dump.misc.index_php import save_IndexPHP
from wikiteam3.dumpgenerator.dump.misc.special_logs import save_SpecialLog
from wikiteam3.dumpgenerator.dump.misc.special_version import save_SpecialVersion
from wikiteam3.dumpgenerator.dump.misc.site_info import save_siteinfo
from wikiteam3.dumpgenerator.dump.xmldump.xml_dump import generate_XML_dump
from wikiteam3.dumpgenerator.dump.xmldump.xml_integrity import check_XML_integrity
from wikiteam3.dumpgenerator.log import log_error
from wikiteam3.utils import url2prefix_from_config, undo_HTML_entities, avoid_WikiMedia_projects
from wikiteam3.utils.ia_checker import any_recent_ia_item_exists
from wikiteam3.utils.util import ALL_DUMPED_MARK, mark_as_done

# From https://stackoverflow.com/a/57008707
class Tee(object):
    def __init__(self, filename):
        self.file = open(filename, 'w', encoding="utf-8")
        self.stdout = sys.stdout

    def __enter__(self):
        sys.stdout = self

    def __exit__(self, exc_type, exc_value, tb):
        sys.stdout = self.stdout
        if exc_type is not None:
            self.file.write(traceback.format_exc())
        self.file.close()

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)

    def flush(self):
        self.file.flush()
        self.stdout.flush()

class DumpGenerator:
    configfilename = "config.json"

    @staticmethod
    def __init__(params=None):
        """Main function"""
        config_filename = DumpGenerator.configfilename
        config, other = get_parameters(params=params)
        avoid_WikiMedia_projects(config=config, other=other)

        with (Tee(other["stdout_log_path"]) if other["stdout_log_path"] is not None else contextlib.nullcontext()):
            print(welcome())
            print("Analysing %s" % (config.api if config.api else config.index))

            # creating path or resuming if desired
            c = 2
            # to avoid concat blabla-2, blabla-2-3, and so on...
            originalpath = config.path
            # do not enter if resume is requested from begining
            while not other["resume"] and os.path.isdir(config.path):
                print('\nWarning!: "%s" path exists' % (config.path))
                reply = ""
                if config.failfast:
                    reply = "yes"
                while reply.lower() not in ["yes", "y", "no", "n"]:
                    reply = input(
                        'There is a dump in "%s", probably incomplete.\nIf you choose resume, to avoid conflicts, some parameters you have chosen in the current session will be ignored\nand the parameters available in "%s/%s" will be loaded.\nDo you want to resume ([yes, y], [no, n])? '
                        % (config.path, config.path, config_filename)
                    )
                if reply.lower() in ["yes", "y"]:
                    if not os.path.isfile("{}/{}".format(config.path, config_filename)):
                        print("No config file found. I can't resume. Aborting.")
                        sys.exit(1)
                    print("You have selected: YES")
                    other["resume"] = True
                    break
                elif reply.lower() in ["no", "n"]:
                    print("You have selected: NO")
                    other["resume"] = False
                config.path = "%s-%d" % (originalpath, c)
                print('Trying to use path "%s"...' % (config.path))
                c += 1

            if other["resume"]:
                print("Loading config file...")
                config = load_config(config=config, config_filename=config_filename)
            else:
                if not other['force'] and any_recent_ia_item_exists(config, days=365):
                    print("A dump of this wiki was uploaded to IA in the last 365 days. Aborting.")
                    sys.exit(88)

                os.mkdir(config.path)
                save_config(config=config, config_filename=config_filename)

            if other["resume"]:
                DumpGenerator.resumePreviousDump(config=config, other=other)
            else:
                DumpGenerator.createNewDump(config=config, other=other)

            save_IndexPHP(config=config, session=other["session"])
            save_SpecialVersion(config=config, session=other["session"])
            save_siteinfo(config=config, session=other["session"])
            mark_as_done(config=config, mark=ALL_DUMPED_MARK)
            bye()
            if other["upload"]:
                print('Calling uploader... (--upload)')
                retcode = subprocess.call([sys.executable, '-m', 'wikiteam3.uploader', config.path] + other["uploader_args"],
                    shell=False)
                if retcode:
                    print(f'--upload: Failed: {retcode}')
                    sys.exit(retcode)
                
                print('--upload: Done')

    @staticmethod
    def createNewDump(config: Config, other: Dict):
        # we do lazy title dumping here :)
        images = []
        print("Trying generating a new dump into a new directory...")
        if config.xml:
            generate_XML_dump(config=config, session=other["session"])
            check_XML_integrity(config=config, session=other["session"])
        if config.images:
            images += Image.get_image_names(config=config, session=other["session"])
            Image.save_image_names(config=config, images=images)
            Image.generate_image_dump(
                config=config, other=other, images=images, session=other["session"]
            )
        if config.logs:
            pass # TODO
            # save_SpecialLog(config=config, session=other["session"])

    @staticmethod
    def resumePreviousDump(config: Config, other: Dict):
        images = []
        print("Resuming previous dump process...")
        if config.xml:

            # checking xml dump
            xml_is_complete = False
            last_xml_title = None
            last_xml_revid = None
            try:
                with FileReadBackwards(
                    "%s/%s-%s-%s.xml"
                    % (
                        config.path,
                        url2prefix_from_config(config=config),
                        config.date,
                        "current" if config.curonly else "history",
                    ),
                    encoding="utf-8",
                ) as frb:
                    for l in frb:
                        if l.strip() == "</mediawiki>":
                            # xml dump is complete
                            xml_is_complete = True
                            break

                        xmlrevid = re.search(r"    <id>([^<]+)</id>", l)
                        if xmlrevid:
                            last_xml_revid = int(xmlrevid.group(1))
                        xmltitle = re.search(r"<title>([^<]+)</title>", l)
                        if xmltitle:
                            last_xml_title = undo_HTML_entities(text=xmltitle.group(1))
                            break

            except:
                pass  # probably file does not exists

            if xml_is_complete:
                print("XML dump was completed in the previous session")
            elif last_xml_title:
                # resuming...
                print('Resuming XML dump from "%s" (revision id %s)' % (last_xml_title, last_xml_revid))
                generate_XML_dump(
                    config=config,
                    session=other["session"],
                    resume=True,
                )
            else:
                # corrupt? only has XML header?
                print("XML is corrupt? Regenerating...")
                generate_XML_dump(config=config, session=other["session"])

        if config.images:
            # load images list
            lastimage = ""
            imagesFilePath = "%s/%s-%s-images.txt" % (config.path, url2prefix_from_config(config=config), config.date)
            if os.path.exists(imagesFilePath):
                f = open(imagesFilePath)
                lines = f.read().splitlines()
                for l in lines:
                    if re.search(r"\t", l):
                        images.append(l.split("\t"))
                if len(lines) == 0: # empty file
                    lastimage = "--EMPTY--"
                if lastimage == "":
                    lastimage = lines[-1].strip()
                if lastimage == "":
                    lastimage = lines[-2].strip()
                f.close()

            if len(images)>0 and len(images[0]) < 5:
                print(
                    "Warning: Detected old images list (images.txt) format.\n"+
                    "You can delete 'images.txt' manually and restart the script."
                )
                sys.exit(1)
            if lastimage == "--END--":
                print("Image list was completed in the previous session")
            else:
                print("Image list is incomplete. Reloading...")
                # do not resume, reload, to avoid inconsistences, deleted images or
                # so
                images = Image.get_image_names(config=config, session=other["session"])
                Image.save_image_names(config=config, images=images)
            # checking images directory
            listdir = []
            if os.path.exists(f"{config.path}/images"):
                listdir = os.listdir(f"{config.path}/images")
            listdir = set(listdir)
            c_images_downloaded = 0
            c_checked = 0
            for filename, url, uploader, size, sha1, timestamp in images:
                if other["filenamelimit"] < len(filename.encode('utf-8')):
                    log_error(
                        config=config, to_stdout=True,
                        text=f"Filename too long(>240 bytes), skipping: {filename}",
                    )
                    continue
                if filename in listdir:
                    c_images_downloaded += 1
                c_checked += 1
                if c_checked % 100000 == 0:
                    print(f"checked {c_checked}/{len(images)} records", end="\r")
            print(f"{len(images)} records in images.txt, {c_images_downloaded} images were saved in the previous session")
            if c_images_downloaded < len(images):
                complete = False
                print("WARNING: Some images were not saved in the previous session")
            else:
                complete = True
            if complete:
                # image dump is complete
                print("Image dump was completed in the previous session")
            else:
                # we resume from previous image, which may be corrupted 
                # by the previous session ctrl-c or abort
                Image.generate_image_dump(
                    config=config,
                    other=other,
                    images=images,
                    session=other["session"],
                )

        if config.logs:
            # fix
            pass



from io import StringIO
import os
from typing import Optional

import lxml.etree
from file_read_backwards import FileReadBackwards


def endsWithNewlines(filename: str) -> int:
    """Returns the number of newlines at the end of file"""

    with FileReadBackwards(filename, encoding="utf-8") as frb:
        newlines = 0
        while frb.readline() == "":
            newlines += 1
    return newlines


def addNewline(filename: str) -> None:
    """Adds a newline to the end of file"""

    print(f"Adding newline to end of {filename}")
    with open(filename, "a", encoding="utf-8") as f:
        f.write("\n")


def truncateXMLDump(filename: str) -> str:
    """Removes incomplete <page> elements from the end of XML dump files"""

    with FileReadBackwards(filename, encoding="utf-8") as frb:
        incomplete_segment: str = ""
        xml_line: str = frb.readline()
        while xml_line and "</title>" not in xml_line:
            incomplete_segment = xml_line + incomplete_segment
            xml_line = frb.readline()
        while xml_line and "</page>" not in xml_line:
            incomplete_segment = xml_line + incomplete_segment
            xml_line = frb.readline()
    incomplete_segment_size = len(incomplete_segment.encode("utf-8"))
    file_size = os.path.getsize(filename)
    if file_size > incomplete_segment_size:
        with open(filename, "r+", encoding="utf-8") as fh:
            fh.truncate(file_size - incomplete_segment_size)
    else:
        print(
            'len(incomplete_segment.encode("utf-8")) returned '
            + str(incomplete_segment_size)
            + ", while os.path.getsize(filename) returned "
            + str(file_size)
            + ", so fh.truncate() would be fh.truncate("
            + str(file_size - incomplete_segment_size)
            + "), which would be illegal. Something is seriously wrong here!"
        )

    # add newline to prevent `</page> <page>` in one line
    if endsWithNewlines(filename) == 0:
        addNewline(filename)
    elif endsWithNewlines(filename) > 1:
        print(
            f"WARNING: {filename} has {endsWithNewlines(filename)} newlines"
        )
    return incomplete_segment

def parse_last_page_chunk(chunk: str) -> Optional[lxml.etree._ElementTree]:
    try:
        parser = lxml.etree.XMLParser(recover=True)
        tree = lxml.etree.parse(StringIO(chunk), parser)
        return tree.getroot()
    except lxml.etree.LxmlError:
        # TODO: log this
        return None


import re

import requests

from wikiteam3.utils import get_random_UserAgent


def get_WikiEngine(url="", session: requests.Session=None) -> str:
    """Returns the wiki engine of a URL, if known"""

    if not session:
        session = requests.Session()  # Create a new session
        session.headers.update({"User-Agent": get_random_UserAgent()})
    r = session.post(url=url, timeout=30)
    if r.status_code == 405 or r.text == "":
        r = session.get(url=url, timeout=120)
    result = r.text

    wikiengine = "Unknown"
    if re.search(
        '(?im)(<meta name="generator" content="DokuWiki)|dokuwiki__site', result
    ):
        wikiengine = "DokuWiki"
    elif re.search(
        '(?im)(alt="Powered by MediaWiki"|<meta name="generator" content="MediaWiki|class="mediawiki)',
        result,
    ):
        wikiengine = "MediaWiki"
    elif re.search(
        '(?im)(>MoinMoin Powered</a>|<option value="LocalSiteMap">)', result
    ):
        wikiengine = "MoinMoin"
    elif re.search(
        "(?im)(twikiCurrentTopicLink|twikiCurrentWebHomeLink|twikiLink)", result
    ):
        wikiengine = "TWiki"
    elif re.search("(?im)(<!--PageHeaderFmt-->)", result):
        wikiengine = "PmWiki"
    elif re.search(
        '(?im)(<meta name="generator" content="PhpWiki|<meta name="PHPWIKI_VERSION)',
        result,
    ):
        wikiengine = "PhpWiki"
    elif re.search(
        r'(?im)(<meta name="generator" content="Tiki Wiki|Powered by <a href="http://(www\.)?tiki\.org"| id="tiki-(top|main)")',
        result,
    ):
        wikiengine = "TikiWiki"
    elif re.search(
        r'(?im)(foswikiNoJs|<meta name="foswiki\.|foswikiTable|foswikiContentFooter)',
        result,
    ):
        wikiengine = "FosWiki"
    elif re.search(r'(?im)(<meta http-equiv="powered by" content="MojoMojo)', result):
        wikiengine = "MojoMojo"
    elif re.search(
        r'(?im)(id="xwiki(content|nav_footer|platformversion|docinfo|maincontainer|data)|/resources/js/xwiki/xwiki|XWiki\.webapppath)',
        result,
    ):
        wikiengine = "XWiki"
    elif re.search(r'(?im)(<meta id="confluence-(base-url|context-path)")', result):
        wikiengine = "Confluence"
    elif re.search(r'(?im)(<meta name="generator" content="Banana Dance)', result):
        wikiengine = "Banana Dance"
    elif re.search(
        r'(?im)(Wheeled by <a class="external-link" href="http://www\.wagn\.org">|<body id="wagn">)',
        result,
    ):
        wikiengine = "Wagn"
    elif re.search(r'(?im)(<meta name="generator" content="MindTouch)', result):
        wikiengine = "MindTouch"  # formerly DekiWiki
    elif re.search(
        r'(?im)(<div class="wikiversion">\s*(<p>)?JSPWiki|xmlns:jspwiki="http://www\.jspwiki\.org")',
        result,
    ):
        wikiengine = "JSPWiki"
    elif re.search(
        r'(?im)(Powered by:?\s*(<br ?/>)?\s*<a href="http://kwiki\.org">|\bKwikiNavigation\b)',
        result,
    ):
        wikiengine = "Kwiki"
    elif re.search(r'(?im)(Powered by <a href="http://www\.anwiki\.com")', result):
        wikiengine = "Anwiki"
    elif re.search(
        '(?im)(<meta name="generator" content="Aneuch|is powered by <em>Aneuch</em>|<!-- start of Aneuch markup -->)',
        result,
    ):
        wikiengine = "Aneuch"
    elif re.search(r'(?im)(<meta name="generator" content="bitweaver)', result):
        wikiengine = "bitweaver"
    elif re.search(r'(?im)(powered by <a href="[^"]*\bzwiki.org(/[^"]*)?">)', result):
        wikiengine = "Zwiki"
    # WakkaWiki forks
    elif re.search(
        r'(?im)(<meta name="generator" content="WikkaWiki|<a class="ext" href="(http://wikka\.jsnx\.com/|http://wikkawiki\.org/)">)',
        result,
    ):
        wikiengine = "WikkaWiki"  # formerly WikkaWakkaWiki
    elif re.search(r'(?im)(<meta name="generator" content="CoMa Wiki)', result):
        wikiengine = "CoMaWiki"
    elif re.search(r'(?im)(Fonctionne avec <a href="http://www\.wikini\.net)', result):
        wikiengine = "WikiNi"
    elif re.search(r'(?im)(Powered by <a href="[^"]*CitiWiki">CitiWiki</a>)', result):
        wikiengine = "CitiWiki"
    elif re.search(
        r'(?im)(Powered by <a href="http://wackowiki\.com/|title="WackoWiki")', result
    ):
        wikiengine = "WackoWiki"
    elif re.search(r'(?im)(Powered by <a href="http://www\.wakkawiki\.com)', result):
        # This may not work for heavily modded/themed installations, e.g.
        # http://operawiki.info/
        wikiengine = "WakkaWiki"
    # Custom wikis used by wiki farms
    elif re.search(r'(?im)(var wikispaces_page|<div class="WikispacesContent)', result):
        wikiengine = "Wikispaces"
    elif re.search(
        r'(?im)(Powered by <a href="http://www\.wikidot\.com">|wikidot-privacy-button-hovertip|javascript:WIKIDOT\.page)',
        result,
    ):
        wikiengine = "Wikidot"
    elif re.search(
        r"(?im)(IS_WETPAINT_USER|wetpaintLoad|WPC-bodyContentContainer)", result
    ):
        wikiengine = "Wetpaint"
    elif re.search(
        '(?im)(<div id="footer-pbwiki">|ws-nav-search|PBinfo *= *{)', result
    ):
        # formerly PBwiki
        wikiengine = "PBworks"
    # if wikiengine == 'Unknown': print (result)

    return wikiengine



import re
from urllib.parse import urlparse, unquote

from slugify import slugify

from wikiteam3.dumpgenerator.config import Config


def url2prefix_from_config(config: Config, ascii_slugify: bool = True):
    """
    Chose a filename/dirname prefix for the dump based on the API url or INDEX url in the config.

    see `url2prefix()` for details
    """

    if url := config.api:
        return url2prefix(url, ascii_slugify=ascii_slugify)
    elif url := config.index:
        return url2prefix(url, ascii_slugify=ascii_slugify)
    else:
        raise ValueError('No URL found in config')


def standardize_url(url: str, strict: bool = True):
    """ 1. strip and unquote url
            > raises ValueError if url contains newline (`\\n` or `\\r`) after stripping
        2. Add `http://` if scheme is missing
            > if `strict` is True, raises ValueError if scheme is missing
        3. Convert domain to IDNA
        4. Remove port `:80` and `:443` if `http://` and `https://` respectively
    """
    # TODO: make step 1,2,4 optional and reversible

    url = url.strip()
    if '\n' in url or '\r' in url:
        raise ValueError('URL contains newline')
    url = unquote(url, encoding='utf-8', errors='strict')

    if not url.startswith('http://') and not url.startswith('https://'):
        if strict:
            raise ValueError(f'HTTP(s) scheme is missing: {url}') 
        print('Warning: URL scheme is missing, assuming http://')
        url = 'http://' + url

    Url = urlparse(url)
    idna_hostname = Url.hostname.encode('idna').decode('utf-8')

    if Url.hostname != idna_hostname:
        print('Converting domain to IDNA: ' + Url.hostname + ' -> ' + idna_hostname)
        url = url.replace(Url.hostname, idna_hostname, 1)
    
    if Url.port == 80 and Url.scheme == 'http':
        print('Removing port 80 from URL')
        url = url.replace(':80', '', 1)
    
    if Url.port == 443 and Url.scheme == 'https':
        print('Removing port 443 from URL')
        url = url.replace(':443', '', 1)

    return url


def url2prefix(url: str, ascii_slugify: bool = True):
    """Convert URL to a valid prefix filename.
    
    1. standardize url (see `standardize_url()`)
    2. remove last slash if exists
    3. truncate to last slash
    4. remove "/any.php" suffix
    5. remove ~ tilde
    6. sulgify the url path if `ascii_slugify` is True
    7. replace port(`:`) with underscore(`_`)
    8. lower case

    """

    url = standardize_url(url)

    r = urlparse(url)

    r_path = r.path

    if r.path.endswith('/'):
        # remove last slash
        # "/abc/123/" -> "/abc/123"
        # "/" -> ""
        r_path = r.path[:-1]
    else: # not r.path.endswith('/')
        # truncate to last slash
        # "/abc/123/edf" -> "/abc/123"
        r_path =  r.path[:r.path.rfind('/')]

    # remove "/any.php" suffix
    r_path = re.sub(r"(/[^/]+\.php)", "", r_path)
    # remove tilde
    r_path = r_path.replace('~', '')
    # sulgify
    _r_paths = r_path.split('/')
    if ascii_slugify:
        _r_paths = [slugify(p, separator='_', allow_unicode=False) for p in _r_paths]
    r_path = '_'.join(_r_paths)

    # replace port with underscore
    r_netloc = r.netloc.replace(':', '_')

    # lower case
    prefix = (r_netloc + r_path).lower()
    assert prefix == prefix.strip('_'), 'prefix contains leading or trailing underscore, please report this bug.'

    return prefix



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wikiteam3',
 'wikiteam3.dumpgenerator',
 'wikiteam3.dumpgenerator.api',
 'wikiteam3.dumpgenerator.cli',
 'wikiteam3.dumpgenerator.dump',
 'wikiteam3.dumpgenerator.dump.image',
 'wikiteam3.dumpgenerator.dump.misc',
 'wikiteam3.dumpgenerator.dump.page',
 'wikiteam3.dumpgenerator.dump.page.xmlexport',
 'wikiteam3.dumpgenerator.dump.page.xmlrev',
 'wikiteam3.dumpgenerator.dump.xmldump',
 'wikiteam3.dumpgenerator.log',
 'wikiteam3.uploader',
 'wikiteam3.utils',
 'wikiteam3.utils.login']

package_data = \
{'': ['*']}

install_requires = \
['file_read_backwards>=3.0.0,<4.0.0',
 'httpx>=0.24.1,<0.25.0',
 'internetarchive>=3.5.0,<4.0.0',
 'lxml>=4.9.2,<5.0.0',
 'mwclient>=0.10.1,<0.11.0',
 'python-slugify>=8.0.1,<9.0.0',
 'requests>=2.31.0,<3.0.0']

entry_points = \
{'console_scripts': ['dumpgenerator = wikiteam3.dumpgenerator:main_deprecated',
                     'wikiteam3dumpgenerator = wikiteam3.dumpgenerator:main',
                     'wikiteam3uploader = wikiteam3.uploader:main']}

setup_kwargs = {
    'name': 'wikiteam3',
    'version': '4.0.16',
    'description': 'Tools for downloading and preserving MediaWikis. We archive MediaWikis, from Wikipedia to tiniest wikis.',
    'long_description': '# `wikiteam3`\n\n<!-- !["MediaWikiArchive.png"](./MediaWikiArchive.png) -->\n<div align=center><img width = \'150\' height =\'150\' src ="./MediaWikiArchive.png"/></div>\n\n> Countless MediaWikis are still waiting to be archived.\n>\n> _Image by [@gledos](https://github.com/gledos/)_\n\n`wikiteam3` is a fork of `mediawiki-scraper`.\n\n## Why we fork mediawiki-scraper\n\nOriginally, mediawiki-scraper was named wikiteam3, but wikiteam upstream (py2 version) suggested that the name should be changed to avoid confusion with the original wikiteam.  \nHalf a year later, we didn\'t see any py3 porting progress in the original wikiteam, and mediawiki-scraper lacks "code" reviewers.  \nSo, we decided to break that suggestion, fork and named it back to wikiteam3, put the code here, and release it to pypi wildly.\n\nEverything still under GPLv3 license.\n\n## Installation\n\n```shell\npip install wikiteam3 --upgrade\n```\n\n## Usage\n\n### Downloading a wiki with complete XML history and images\n\n```bash\nwikiteam3dumpgenerator http://wiki.domain.org --xml --images\n```\n\n### Manually specifying `api.php` and/or `index.php`\n\nIf the script can\'t find itself the `api.php` and/or `index.php` paths, then you can provide them:\n\n```bash\nwikiteam3dumpgenerator --api http://wiki.domain.org/w/api.php --xml --images\n```\n\n```bash\nwikiteam3dumpgenerator --api http://wiki.domain.org/w/api.php --index http://wiki.domain.org/w/index.php \\\n    --xml --images\n```\n\nIf you only want the XML histories, just use `--xml`. For only the images, just `--images`. For only the current version of every page, `--xml --curonly`.\n\n### Resuming an incomplete dump\n\n```bash\nwikiteam3dumpgenerator \\\n    --api http://wiki.domain.org/w/api.php --xml --images --resume --path /path/to/incomplete-dump\n```\n\nIn the above example, `--path` is only necessary if the download path is not the default.\n\n`wikiteam3dumpgenerator` will also ask you if you want to resume if it finds an incomplete dump in the path where it is downloading.\n\n## Using `wikiteam3uploader`\n\nTODO: ...\n\n### Requirements\n\n- unbinded port 62954\n- 3GB+ RAM (~2.56GB for commpressing)\n- 64-bit OS (required by 2G wlog size)\n\n- 7z (`7z-full` with lzma2)\n- zstd 1.5.5+ (recommended), v1.5.0-v1.5.4(DO NOT USE), 1.4.8 (minimum)\n\n## Checking dump integrity\n\nTODO: xml2titles.py\n\nIf you want to check the XML dump integrity, type this into your command line to count title, page and revision XML tags:\n\n```bash\ngrep -E \'<title(.*?)>\' *.xml -c;grep -E \'<page(.*?)>\' *.xml -c;grep \\\n    "</page>" *.xml -c;grep -E \'<revision(.*?)>\' *.xml -c;grep "</revision>" *.xml -c\n```\n  \nYou should see something similar to this (not the actual numbers) - the first three numbers should be the same and the last two should be the same as each other:\n\n```bash\n580\n580\n580\n5677\n5677\n```\n\nIf your first three numbers or your last two numbers are different, then, your XML dump is corrupt (it contains one or more unfinished ```</page>``` or ```</revision>```). This is not common in small wikis, but large or very large wikis may fail at this due to truncated XML pages while exporting and merging. The solution is to remove the XML dump and re-download, a bit boring, and it can fail again.\n\n## Contributors\n\n**WikiTeam** is the [Archive Team](http://www.archiveteam.org) [[GitHub](https://github.com/ArchiveTeam)] subcommittee on wikis.\nIt was founded and originally developed by [Emilio J. Rodríguez-Posada](https://github.com/emijrp), a Wikipedia veteran editor and amateur archivist. Thanks to people who have helped, especially to: [Federico Leva](https://github.com/nemobis), [Alex Buie](https://github.com/ab2525), [Scott Boyd](http://www.sdboyd56.com), [Hydriz](https://github.com/Hydriz), Platonides, Ian McEwen, [Mike Dupont](https://github.com/h4ck3rm1k3), [balr0g](https://github.com/balr0g) and [PiRSquared17](https://github.com/PiRSquared17).\n\n**Mediawiki-Scraper** The Python 3 initiative is currently being led by [Elsie Hupp](https://github.com/elsiehupp), with contributions from [Victor Gambier](https://github.com/vgambier), [Thomas Karcher](https://github.com/t-karcher), [Janet Cobb](https://github.com/randomnetcat), [yzqzss](https://github.com/yzqzss), [NyaMisty](https://github.com/NyaMisty) and [Rob Kam](https://github.com/robkam)\n\n**WikiTeam3** None yet.\n',
    'author': 'yzqzss',
    'author_email': 'yzqzss@yandex.com',
    'maintainer': 'yzqzss',
    'maintainer_email': 'yzqzss@yandex.com',
    'url': 'https://github.com/saveweb/wikiteam3',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import dataclasses
import json
from typing import List


def _dataclass_from_dict(klass_or_obj, d: dict):
    if isinstance(klass_or_obj, type): # klass
        ret = klass_or_obj()
    else:
        ret = klass_or_obj
    for k,v in d.items():
        if hasattr(ret, k):
            setattr(ret, k, v)
    return ret

'''
config = {
        "curonly": args.curonly,
        "date": datetime.datetime.now().strftime("%Y%m%d"),
        "api": api,
        "failfast": args.failfast,
        "http_method": "POST",
        "index": index,
        "images": args.images,
        "logs": False,
        "xml": args.xml,
        "xmlrevisions": args.xmlrevisions,
        "namespaces": namespaces,
        "exnamespaces": exnamespaces,
        "path": args.path and os.path.normpath(args.path) or "",
        "cookies": args.cookies or "",
        "delay": args.delay,
        "retries": int(args.retries),
    }
'''
@dataclasses.dataclass
class Config:
    def asdict(self):
        return dataclasses.asdict(self)

    # General params
    delay: float = 0.0
    retries: int = 0
    path: str = ''
    logs: bool = False
    date: str = False

    # URL params
    index: str = ''
    api: str = ''

    # Download params
    xml: bool = False
    curonly: bool = False
    xmlapiexport: bool = False
    xmlrevisions: bool = False
    xmlrevisions_page: bool = False
    images: bool = False
    namespaces: List[int] = None
    exnamespaces: List[int] = None

    api_chunksize: int = 0  # arvlimit, ailimit, etc
    export: str = '' # Special:Export page name
    http_method: str = ''

    # Meta info params
    failfast: bool = False

    templates: bool = False

def new_config(configDict) -> Config:
    return _dataclass_from_dict(Config, configDict)

def load_config(config: Config=None, config_filename=""):
    """Load config file"""

    config_dict = dataclasses.asdict(config)

    if config.path:
        try:
            with open(f"{config.path}/{config_filename}", encoding="utf-8") as infile:
                config_dict.update(json.load(infile))
            return new_config(config_dict)
        except FileNotFoundError:
            pass

    raise RuntimeError("There is no config file. we can't resume.")

def save_config(config: Config, config_filename: str):
    """Save config file"""

    with open(f"{config.path}/{config_filename}", "w", encoding="utf-8") as outfile:
        json.dump(dataclasses.asdict(config), outfile, indent=4, sort_keys=True)


from wikiteam3.dumpgenerator.config import Config
from .page_xml_api import getXMLPageWithApi
from .page_xml_export import getXMLPageWithExport


def get_XML_page(config: Config = None, title="", verbose=True, session=None):
    if config.xmlapiexport:
        return getXMLPageWithApi(config=config, title=title, verbose=verbose, session=session)
    else:
        return getXMLPageWithExport(config=config, title=title, verbose=verbose, session=session)



""" Provide login functions """

import requests
import time

from wikiteam3.utils.login.api import bot_login, client_login, fetch_login_token
from wikiteam3.utils.login.index import index_login


def uniLogin(api: str = '', index: str = '' ,session: requests.Session = requests.Session(), username: str = '', password: str = ''):
    """ Try to login to a wiki using various methods.\n
    Return `session` if success, else return `None`.\n
    Try: `cilent login (api) => bot login (api) => index login (index)` """

    if (not api and not index) or (not username or not password):
        raise ValueError('uniLogin: api or index or username or password is empty')

    if api:
        print("Trying to log in to the wiki using clientLogin... (MW 1.27+)")
        _session = client_login(api=api, session=session, username=username, password=password)
        if _session:
            return _session
        time.sleep(5)

        print("Trying to log in to the wiki using botLogin... (MW 1.27+)")
        _session = bot_login(api=api, session=session, username=username, password=password)
        if _session:
            return _session
        time.sleep(5)

    if index:
        print("Trying to log in to the wiki using indexLogin... (generic)")
        _session = index_login(index=index, session=session, username=username, password=password)
        if _session:
            return _session

    return None



import datetime

from wikiteam3.dumpgenerator.config import Config

def log_error(config: Config, to_stdout=False , text="") -> None:
    """Log error in errors.log"""
    if text:
        with open(f"{config.path}/errors.log", "a", encoding="utf-8") as outfile:
            output = "{}: {}\n".format(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                text,
            )
            outfile.write(output)
    if to_stdout:
        print(text)
